"""Test base.py"""


from chromadol.base import ChromaClient


def test_simple():
    """A simple test of the ChromaClient and ChromaCollection classes."""

    import tempfile, os

    with tempfile.TemporaryDirectory() as temp_dir:
        tempdir = os.path.join(temp_dir, 'chromadol_test')
        os.makedirs(tempdir)
    client = ChromaClient(tempdir)

    # removing all contents of client to be able to run a test on a clean slate
    for k in client:
        del client[k]
    assert list(client) == []
    collection = client['chromadol_test']
    # note that just accessing the collection creates it (by default)
    assert list(client) == ['chromadol_test']
    assert list(collection) == []
    assert len(collection) == 0

    # chromadb is designed to operate on multiple documents at once, so
    # specifying it's keys and contents (and any extras) list this:
    collection[['piece', 'of']] = {
        'documents': ['contents for piece', 'contents for of'],
        'metadatas': [{'author': 'me'}, {'author': 'you'}],
    }

    # Now we have two documents in the collection:

    assert len(collection) == 2

    # Note, though, that the order of the documents is not guaranteed.

    assert sorted(collection) == ['of', 'piece']

    assert collection['piece'] == {
        'ids': ['piece'],
        'embeddings': None,
        'metadatas': [{'author': 'me'}],
        'documents': ['contents for piece'],
        'uris': None,
        'data': None,
    }

    assert collection['of'] == {
        'ids': ['of'],
        'embeddings': None,
        'metadatas': [{'author': 'you'}],
        'documents': ['contents for of'],
        'uris': None,
        'data': None,
    }

    # You can also read multiple documents at once.
    # But note that the order of the documents is not guaranteed.
    assert collection[['piece', 'of']] == collection[['of', 'piece']]

    # But you can read or write one document at a time too.
    collection['cake'] = {
        'documents': 'contents for cake',
    }
    assert set(collection) == {'piece', 'of', 'cake'}
    assert collection['cake'] == {
        'ids': ['cake'],
        'embeddings': None,
        'metadatas': [None],
        'documents': ['contents for cake'],
        'uris': None,
        'data': None,
    }

    # In fact, see that if you only want to specify the "documents" part of the information,
    collection['cake'] = 'a different cake'
    assert collection['cake'] == {
        'ids': ['cake'],
        'embeddings': None,
        'metadatas': [None],
        'documents': ['a different cake'],
        'uris': None,
        'data': None,
    }

    # The `collection` instance is not only dict-like, but also list-like in the
    # sense that it has an `.append` and an `.extend` method.

    assert len(collection) == 3
    collection.extend(['two documents', 'specified without keys'])
    assert len(collection) == 5



"""Base objects for chromadol."""

from typing import MutableMapping, Union
from functools import cached_property
from dol.appendable import appendable, mk_item2kv_for

from chromadb import Client, PersistentClient, GetResult

dflt_create_collection_kwargs = dict()


class ChromaClient(MutableMapping):
    def __init__(self, client=None, *, get_or_create=True, **create_collection_kwargs):
        """
        Initializes the reader with a chromadb Client instance.

        :param client: An instance of chromadb.Client.
        """
        if client is None:
            client = Client()
        elif isinstance(client, str):
            client = PersistentClient(client)
        self.client = client
        self._create_collection_kwargs_for_getitem = dict(
            create_collection_kwargs, get_or_create=get_or_create
        )
        self._create_collection_kwargs_for_setitem = create_collection_kwargs

    def __iter__(self):
        """
        Iterates over the names of collections in the chromadb Client.
        """
        return (obj.name for obj in self.client.list_collections())

    def __getitem__(self, k: str):
        """
        Retrieves a collection by its name. Creates the collection if it doesn't exist.

        :param k: The name of the collection to retrieve.
        :return: The collection object.
        """
        return ChromaCollection(
            self.client.create_collection(
                k, **self._create_collection_kwargs_for_getitem
            )
        )

    def __setitem__(self, k: str, v: dict):
        """
        Creates or updates a collection.

        :param k: The name of the collection.
        :param v: a dict that will be used to populate the collection via .add_documents(**v)
        """
        # Implementation depends on how collections are created or updated in chromadb
        # Example:
        collection = self.client.create_collection(
            name=k, **self._create_collection_kwargs_for_setitem
        )
        if v:
            collection.add_documents(**v)

    def __delitem__(self, k: str):
        """
        Deletes a collection.

        :param k: The name of the collection to delete.
        """
        self.client.delete_collection(k)

    def __len__(self):
        """
        Returns the number of collections in the client.
        """
        return len(self.client.list_collections())

    def __contains__(self, k):
        """
        Returns True if the client has a collection with the given name.
        """
        existing_names = set(self)
        return k in existing_names

    def clear(self):
        """
        This method is here, in fact, to disable the clear method, that would
        otherwise be inherited from MutableMapping.
        It's existence is too dangerous, as it would delete all collections in the
        client.
        If you want to actually delete all collections in the client, do so explicitly
        by iterating over the client and deleting each collection, as such:

        >>> for k in chroma_client_instance:  # doctest: +SKIP
        ...     del chroma_client_instance[k]

        """
        raise NotImplementedError('Disabled for safety reasons.')


def int_string(x):
    return str(int(x))


item2kv = mk_item2kv_for.utc_key(factor=1e9, time_postproc=int_string)


@appendable(item2kv=item2kv)
class ChromaCollection(MutableMapping):
    def __init__(self, collection):
        """
        Initializes the store with a chromadb Collection instance.

        :param collection: An instance of chromadb.Collection.
        """
        self.collection = collection

    @property
    def _ids(self):
        collection_elements = self.collection.get()
        return collection_elements['ids']

    def __iter__(self):
        return iter(self._ids)

    def __getitem__(self, k: str) -> GetResult:
        return self.collection.get(k)

    def __len__(self):
        return self.collection.count()

    def __contains__(self, k):
        try:
            self.collection.get(k)
            return True
        except KeyError:
            return False

    def __setitem__(self, k: str, v: Union[dict, str]):
        if isinstance(v, str):
            v = {'documents': [v]}
        self.collection.upsert(k, **v)

    def __delitem__(self, k: str):
        self.collection.delete(k)



"""Utils for chromadol."""

from functools import partial
from concurrent.futures import ThreadPoolExecutor


def vectorize(func, iterable=None, *, max_workers: int = 1):
    """Like builtin map, but returns a list,
    and if iterable is None, returns a partial function that can directly be applied
    to an iterable.
    This is useful, for instance, for making a data loader from any single-uri loader.

    Example:
    >>> vectorize(lambda x: x**2, [1,2,3])
    [1, 4, 9]
    >>> vectorized_square = vectorize(lambda x: x**2)
    >>> vectorized_square([1,2,3])
    [1, 4, 9]
    """
    if iterable is None:
        return partial(vectorize, func, max_workers=max_workers)
    if max_workers == 1:
        return list(map(func, iterable))
    else:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            return list(executor.map(func, iterable))



"""Data loaders for ChromaDB."""
import multiprocessing
from typing import Optional, Sequence, List, Union
from operator import attrgetter
from chromadb.api.types import URI, DataLoader
from chromadol.util import vectorize

# --------------------------- Examples of loaders ---------------------------

FileContents = Union[str, bytes]  # a type for the contents of a file


def load_text(filepath: str) -> str:
    with open(filepath, 'r') as f:
        return f.read()


def load_bytes(filepath: str) -> bytes:
    with open(filepath, 'rb') as f:
        return f.read()


def url_to_contents(
    url: str, content_extractor=attrgetter('text'), *, params=None, **kwargs
) -> FileContents:
    import requests

    response = requests.get(url, params=params, **kwargs)
    response.raise_for_status()
    return content_extractor(response)


def pdf_file_text(
    filepath: str, *, page_break_delim='---------------------------'
) -> str:
    from pypdf import PdfReader  # pip install pypdf

    return page_break_delim.join(
        page.extract_text() for page in PdfReader(filepath).pages
    )


# --------------------------- FileLoader ---------------------------


class FileLoader(DataLoader[List[Optional[FileContents]]]):
    """A DataLoader that loads a list of text files from a list of URIs.

    By default, it loads text files from local files, given URIs that are full file paths.
    You can specify a prefix thought (usually used to specify a root directory),
    or a suffix (usually used to specify a file extension).

    Further, you can specify a different `loader`, e.g. to load files from a remote URL,
    or to load binary files, or to load text from pdf files, or from S3, or a database, etc.

    Example:

    >>> import chromadb
    >>> rootdir = chromadb.__path__[0] + '/'
    >>> file_loader_1 = FileLoader(prefix=rootdir)
    >>> file_contents_1 = file_loader_1(['__init__.py', 'types.py'])
    >>> len(file_contents_1)
    2
    >>> 'Embeddings' in file_contents_1[0]  # i.e. __init__.py contains the word 'Embeddings'
    True
    >>> 'from typing import' in file_contents_1[1]  # i.e. types.py contains the phrase 'from typing import'
    True

    """

    def __init__(
        self,
        loader=load_text,
        *,
        prefix: str = '',
        suffix: str = '',
        max_workers: int = multiprocessing.cpu_count(),
    ) -> None:
        """
        Args:
            loader: A function that takes a (single) URI and returns its contents
            prefix: A string to prepend to the URI
            suffix: A string to append to the URI
            max_workers: The maximum number of threads to use when loading the URIs
        """
        self._loader = loader
        self._prefix = prefix
        self._suffix = suffix
        self._max_workers = max_workers

    def _load_file(self, uri: Optional[URI]) -> Optional[FileContents]:
        if uri is None:
            return None
        return self._loader(f'{self._prefix}{uri}{self._suffix}')

    def __call__(self, uris: Sequence[Optional[URI]]) -> List[Optional[FileContents]]:
        if isinstance(uris, str):
            # To avoid a common mistake, we cast a string to a list of containing it
            uris = [uris]
        return vectorize(self._load_file, uris, max_workers=self._max_workers)


# add a few loaders as attributes, for convenience
FileLoader.load_text = load_text
FileLoader.load_bytes = load_bytes
FileLoader.url_to_contents = url_to_contents
FileLoader.pdf_file_text = pdf_file_text


def test_file_loader():
    # Here, we'll use the dfault loader, which loads text from local files
    # We'll use the rootdir of the chromadb package as our root directory
    import chromadol

    rootdir = chromadol.__path__[0] + '/'
    file_loader_1 = FileLoader(prefix=rootdir)

    file_contents_1 = file_loader_1(['__init__.py', 'base.py'])
    assert len(file_contents_1) == 2

    # Note: The following two asserts seem faily robust, but still, they assume something
    # about the contents of __init__.py, and types.py, which could change in the future.
    assert 'ChromaCollection' in file_contents_1[0]
    assert 'chromadb' in file_contents_1[1]

    # See that we could also decide that keys should assume the `.py` extension implicitly:
    file_loader_2 = FileLoader(prefix=rootdir, suffix='.py')

    file_contents_2 = file_loader_2(['__init__', 'base'])
    assert sorted(file_contents_1) == sorted(file_contents_2)

    # Now lets use a different loader: One that returns contents from a remote URL
    file_loader_3 = FileLoader(loader=FileLoader.url_to_contents)
    url = 'https://github.com/chroma-core/chroma/issues/1606'
    contents = file_loader_3(url)[0]
    'vectorizer' in contents  # this issue mentions the term "vectorizer"



"""Data Object Layer (DOL) for ChromaDB

Example usage:

To make a `ChromaClient` DOL, you can specify a `chromadb` `Client`, `PersistentClient` (etc.) 
instance, or specify a string (which will be interpreted as a path to a directory to
save the data to in a `PersistentClient` instance).

>>> from chromadol import ChromaClient
>>> import tempfile, os 
>>> with tempfile.TemporaryDirectory() as temp_dir:
...     tempdir = os.path.join(temp_dir, "chromadol_test")
...     os.makedirs(tempdir)
>>> client = ChromaClient(tempdir)

Removing all contents of client to be able to run a test on a clean slate

>>> for k in client:
...     del client[k]
...

There's nothing yet:

>>> list(client)
[]

Now let's "get" a collection. 

>>> collection = client['chromadol_test']

Note that just accessing the collection creates it (by default)


>>> list(client)
['chromadol_test']

Here's nothing in the collection yet:

>>> list(collection)
[]

So let's write something.
Note that `chromadb` is designed to operate on multiple documents at once, 
so the "chromadb-natural" way of specifying it's keys and contents (and any extras) 
would be like this:

>>> collection[['piece', 'of']] = {
...     'documents': ['contents for piece', 'contents for of'],
...     'metadatas': [{'author': 'me'}, {'author': 'you'}],
... }

Now we have two documents in the collection:

>>> len(collection)
2

Note, though, that the order of the documents is not guaranteed.

>>> sorted(collection)
['of', 'piece']

>>> assert collection['piece'] == {
...     'ids': ['piece'],
...     'embeddings': None,
...     'metadatas': [{'author': 'me'}],
...     'documents': ['contents for piece'],
...     'uris': None,
...     'data': None
... }

>>> assert collection['of'] == {
...     'ids': ['of'],
...     'embeddings': None,
...     'metadatas': [{'author': 'you'}],
...     'documents': ['contents for of'],
...     'uris': None,
...     'data': None
... }

You can also read multiple documents at once.
But note that the order of the documents is not guaranteed.

>>> collection[['piece', 'of']] == collection[['of', 'piece']]
True

You can read or write one document at a time too.

>>> collection['cake'] = {
...     "documents": "contents for cake",
... }
>>> sorted(collection)  # sorting because order is not guaranteed
['cake', 'of', 'piece']
>>> assert collection['cake'] == {
...     'ids': ['cake'],
...     'embeddings': None,
...     'metadatas': [None],
...     'documents': ['contents for cake'],
...     'uris': None,
...     'data': None,
... }

In fact, see that if you only want to specify the "documents" part of the information,
you can just write a string instead of a dictionary:

>>> collection['cake'] = 'a different cake'
>>> assert collection['cake'] == {
...     'ids': ['cake'],
...     'embeddings': None,
...     'metadatas': [None],
...     'documents': ['a different cake'],
...     'uris': None,
...     'data': None,
... }

The `collection` instance is not only dict-like, but also list-like in the 
sense that it has an `.append` and an `.extend` method.

>>> len(collection)
3
>>> collection.extend(['two documents', 'specified without keys'])
>>> len(collection)
5

See that the two documents were added to the collection, and that they were assigned
keys automatically:

>>> list(collection)  # doctest: +SKIP
['piece', 'of', 'cake', '1704294539590875904', '1704294539631522048']

"""

from chromadol.base import ChromaCollection, ChromaClient
from chromadol.data_loaders import FileLoader
from chromadol.util import vectorize
